#!/usr/bin/env python3

import argparse

import bacolonyzer
from bacolonyzer import commands


# Entrypoint from the command line.
def run():
    # Create main inputs parser.
    parser = argparse.ArgumentParser(
        description="""Commands and tools for the analysis of QFA images.""")
    # Add version argument.
    parser.add_argument(
        "-v",
        "--version",
        help="show current version",
        action='version',
        version="BaColonyzer {}".format(bacolonyzer.__version__))

    # Register all possible commands of BaColonyzer to the main inputs parser.
    subparsers = parser.add_subparsers()
    commands.AnalyseCommand().register_parser(subparsers)
    commands.RenameImagesCommand().register_parser(subparsers)

    # Actually parsing the inputs given by the user.
    options = parser.parse_args()
    # Execute the run functions depending on the command given on the terminal.
    options.func(options)



"""Functions to get the agar and spot colors, to locate the grid, and to
calibrate the image colors according to a black and white reference picture.
"""

import itertools
import os

import cv2
import numpy as np
from scipy.signal import find_peaks
from tqdm import tqdm


def get_agar_spot_color(im):
    """Get the color value of the agar and the spot.
    Obtain the values by checking the largest two peaks in the color histogram.
    """
    im_ = np.asarray(im).ravel()

    # Compute histogram of intensities
    bincnt = np.bincount(im_)

    # Find peaks in histogram
    peaks, _ = find_peaks(bincnt, width=3)
    vals = bincnt[peaks]
    idx = np.argsort(vals)[::-1]

    # Take first peak: this will be the agar, as it's the most abundant color
    color_agar = peaks[idx[0]]

    # Force the color of the spot to be higher than the color of the agar.
    second_idx = idx[idx > idx[0]]
    if len(second_idx) == 0:
        color_spot = 0.7 * im_.max()
    else:
        color_spot = peaks[second_idx[0]]

    # Set limits that are resonable to avoid noise affecting results
    lower_lim_spots = 2 * color_agar
    upper_lim_spots = 0.7 * im_.max()
    if color_spot < lower_lim_spots:
        color_spot = min(lower_lim_spots, upper_lim_spots)
    elif color_spot > upper_lim_spots:
        color_spot = max(lower_lim_spots, upper_lim_spots)

    return int(color_agar), int(color_spot)


def get_position_grid(im, nrow, ncol, frac):
    """Get the position of the grid in the image.
    Use an artificial pattern and the normalized squared difference to find it.
    Return the position of the grid and the size of the scaled the pattern.
    """
    # Find color values of agar and spots
    color_agar, color_spot = get_agar_spot_color(im)
    # Size of the given image.
    h, w = im.shape

    # Create the pattern of circles following the grid structure.
    rpix = int(h / nrow)  # Set random size of window that will later be scaled
    rspot = int((rpix * 0.5) / 2)  # Set random size of spot radius to scale
    pattern = np.ones((nrow * rpix, ncol * rpix), dtype=np.uint8) * color_agar
    for i, j in itertools.product(range(nrow), range(ncol)):
        cv2.circle(pattern,
                   (j * rpix + int(rpix / 2), i * rpix + int(rpix / 2)), rspot,
                   color_spot, -1)

    # Check for structure between frac and 100% of the total image size.
    min_fraction, max_fraction = frac, 1.0
    # Check for structure using iterations by increasing 0.2% of the image
    iterations = int((max_fraction - min_fraction) * 100 / 0.2)

    # Store the info about the best scale where pattern fits the image.
    found = None
    for fraction in tqdm(
            np.linspace(min_fraction, max_fraction, iterations),
            desc='Fitting pattern...'):
        # Scale the pattern to the fraction of the given image.
        pat_w = int(w * fraction)
        pat_h = int(pat_w * nrow / ncol)
        pattern_scaled = cv2.resize(pattern, (pat_w, pat_h))
        # Match the scaled pattern with the image.
        res = cv2.matchTemplate(im, pattern_scaled, cv2.TM_SQDIFF_NORMED)
        # The function gives the map of squared difference, so optimal location
        # will be the one with minimum value.
        min_val, max_val, min_loc, max_loc = cv2.minMaxLoc(res)
        if found is None or min_val < found[0]:
            found = (min_val, min_loc, pat_h, pat_w)
    return found


def calibration_maxmin(ref_img):
    """Obtain minimum and maximum intensity values captured by the camera by
    using a reference image. This will be used to calibrate the final results.
    """

    # Make sure that reference image exists
    if os.path.isfile(ref_img):
        # Open reference image and get all the intensity values
        reference_image = cv2.imread(ref_img, cv2.IMREAD_GRAYSCALE)

        # # Take 1% and 99% quantiles of the reference image to remove noise
        quants = np.quantile(reference_image.ravel(), [0.01, 0.99])
        reference_image = np.clip(reference_image, *quants)

        min_ref = reference_image.min()
        max_ref = reference_image.max()
    else:
        logger.info(
            """Reference image not found. Calibration not performed.""")
        min_ref, max_ref = 0, 255
    return (min_ref, max_ref)



"""Definition of all commands available in BaColonyzer."""
import logging

from bacolonyzer import analysis, filesystem, utils
from bacolonyzer.commands import abstract

logger = logging.getLogger(__name__)


class AnalyseCommand(abstract.AbstractCommand):

    _SUBCOMMAND = 'analyse'
    _DESCRIPTION = """\
    Analyse timeseries of QFA images: locate cultures on
    plate, segment image into agar and cells, apply lighting correction,
    and generate output files for each image.
    """

    def register_arguments(self, parser):
        parser.add_argument(
            "-d",
            "--directory",
            type=str,
            help="""Directory in which to search for image files that need to be
            analysed.
            Default: current directory.""",
            default=".")
        parser.add_argument(
            "-c",
            "--light_correction_off",
            help="""Disables lighting correction between images.
            Default: Light correction is enabled.""",
            action="store_false")
        parser.add_argument(
            "-r",
            "--reference_image",
            type=str,
            help="""Path to a reference image that shows a white and a black
            paper next to each other. Important: this picture must be taken
            using the same camera settings as the image series. This is used
            to calibrate the final results and might be useful to compare
            different experiments.
            Default: No reference image is used.""",
            default="")
        parser.add_argument(
            "-q",
            "--quiet",
            help="""Suppresses messages printed during the analysis.
            Default: show messages.""",
            action="store_true")
        parser.add_argument(
            "-e",
            "--endpoint",
            help="""Analyses only the final image in the series.
            It is useful to test single images.
            Default: False. Analyse all images in the directory.""",
            action="store_true")
        parser.add_argument(
            "-g",
            "--grid_format",
            type=str,
            nargs='+',
            help="""Specifies rectangular grid format.
            Important: specify number of rows and number of columns, in this order
            (e.g. -g 8x12 or -g 8 12).
            Default: 8x12.""",
            default=['8x12'])
        parser.add_argument(
            "-f",
            "--fraction",
            type=utils.range_float,
            help="""Specifies the minimum fraction of the image that corresponds
            to the grid. Adjust it if grid occupies a small part of the total image.
            Default: 0.8.""",
            default=0.8)

    def run(self, args):
        # Setup logger
        if args.quiet:
            logging.basicConfig(format="%(message)s", level=logging.INFO)
        else:
            logging.basicConfig(format="%(message)s", level=logging.DEBUG)

        # Print information of inputs to users.
        utils.summarise(args)
        filesystem.reference_info(args.reference_image)

        # Get working directory.
        fdir = filesystem.get_directory(args.directory)
        nrow, ncol = utils.get_grid_format(args.grid_format)

        # Create needed directories to save outputs.
        filesystem.arrange_directories(fdir)
        # Obtain list of images to analyse.
        imanalyse = filesystem.get_images(fdir, args.endpoint,
                                          args.reference_image)

        # Perform main logic.
        analysis.analyse_timeseries_qfa(
            imanalyse,
            nrow,
            ncol,
            fdir,
            light_correction=args.light_correction_off,
            fraction=args.fraction,
            reference_image=args.reference_image)

        logger.info("No more images to analyse. I'm done")



import datetime
import glob
import os
import time

from bacolonyzer.commands import abstract
from PIL import Image


class RenameImagesCommand(abstract.AbstractCommand):

    _SUBCOMMAND = 'rename_images'
    _DESCRIPTION = """\
    Rename image files according to date time format. By default it only shows
    the renaming of the image files, without applying the modification. Use the
    flag --no_dry_run to apply the renaming.
    """

    def register_arguments(self, parser):
        parser.add_argument(
            "-d",
            "--directory",
            help="""Directory where the images are stored.
            Default: current directory.""",
            default=".")
        parser.add_argument(
            "-g",
            "--glob",
            help="""Glob to match all images in the given directory.
            Default: *.jpg (matches all jpg files).""",
            default="*.[Jj][Pp][Gg]")
        parser.add_argument(
            "-p",
            "--prefix",
            help="""File prefix for the ranaming. Default: QFAxxxxxxxxxx_,
            where xxxxxxxxxx is a random number.""",
            default="QFA%s_" % (int(time.time())))
        parser.add_argument(
            "-m",
            "--no_read_metadata",
            help="""By default, image files will be renamed using the date and
            time information retrieved from images metadata. If you specify
            this flag, the renaming will take place according to
            parameters -s and -i.""",
            action="store_true")
        parser.add_argument(
            "-s",
            "--start_time",
            help="""Initial time for files renaming.
            Required format: 'YYYY-MM-DD hh:mm'. Default: now.""",
            type=lambda s: datetime.datetime.strptime(s, '%Y-%m-%d %H:%M'),
            default=datetime.datetime.now())
        parser.add_argument(
            "-i",
            "--interval",
            help="Time interval between files in minutes. Default: 30.",
            type=int,
            default=30)
        parser.add_argument(
            "--no_dry_run",
            help="""By default the change of names won't take place.
            If you specify this flag the renaming will take place.""",
            action="store_true")

    def run(self, args):
        # Get list of files to rename.
        file_names = glob.glob(os.path.join(args.directory, args.glob))
        file_names.sort()

        if not file_names:
            raise RuntimeError(
                'No file names found with the glob expression: {}'.format(
                    args.glob))
        print('Found {} files to rename.'.format(len(file_names)))

        # Define starting date time and time interval according to -s and -i
        current_datetime = args.start_time
        delta_time = datetime.timedelta(minutes=args.interval)

        for file_path in file_names:
            # Get extension of the files
            extension = os.path.splitext(file_names[0])[1]

            if args.no_read_metadata:
                # Use default or imputed values of date and time
                date_time = current_datetime.strftime("%Y-%m-%d_%H-%M-%S")
                current_datetime += delta_time
            else:
                # Obtain date and time from image metadata
                exif_data = Image.open(file_path)._getexif()
                if not exif_data:
                    raise RuntimeError(
                        '''Metadata is not available from these images.
                        Please, use parameters -m, -s and -i
                        to properly rename your images.''')
                else:
                    exif_data = exif_data[36867]
                    d_t = datetime.datetime.strptime(exif_data,
                                                     '%Y:%m:%d %H:%M:%S')
                    date_time = d_t.strftime("%Y-%m-%d_%H-%M-%S")

            # Create new file name
            new_file_name = args.prefix + date_time + extension
            new_path = os.path.join(os.path.dirname(file_path), new_file_name)

            # Show the renaming to the user.
            print("{} -> {}".format(file_path, new_path))
            # If specified by users, perform the actual renaming of files.
            if args.no_dry_run:
                os.rename(file_path, new_path)

        # Print information to users
        if args.no_dry_run:
            print("All files renamed!")



#!/usr/bin/env python3

from distutils.core import setup

import setuptools

import bacolonyzer

_DESCRIPTION = """\
BaColonyzer is a software to quantify the cell density of bacterial cultures
from timecourse pictures. It uses image analysis tools to determine the fitness
of spotted cultures of bacteria grown on solid agar.
"""

setup(
    name='bacolonyzer',
    version=bacolonyzer.__version__,
    description='Analyse timeseries of QFA images',
    author='Judith Bergada',
    author_email='judithbergada@gmail.com',
    url='https://github.com/judithbergada/bacolonyzer',
    packages=setuptools.find_packages(),
    long_description=_DESCRIPTION,
    license='MIT',
    entry_points={
        'console_scripts': ['bacolonyzer=bacolonyzer.commands.entrypoint:run'],
    },
    python_requires='>3.4, <4',
    install_requires=[
        'numpy>=1.13.0',
        'opencv-python>=4.0.0.21',
        'pandas>=0.21.1',
        'scipy>=1.0.0',
        'tqdm>=4.28.0',
        'Pillow>=6.0.0',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Image Recognition',
    ],
)



"""Functions to set the inputs that will be used.
"""

import logging

logger = logging.getLogger(__name__)


# Set input parameters for analysis
def summarise(args):
    """Read user input, set up flags for analysis, report on
    options chosen and find files to be analysed."""

    # Get variables nrow and ncol according to rows and columns provided
    nrow, ncol = get_grid_format(args.grid_format)

    # If required by the user, summarise basic input information
    logger.debug("")
    logger.debug("Summary of inputs:")
    logger.debug("Grid:")
    logger.debug("Expecting %s rows and %s columns on plate.", nrow, ncol)
    logger.debug("Searching for colony locations automatically.")
    logger.debug("Assuming that grid occupies at least {:.1f}%".format(
        args.fraction * 100))

    logger.debug("Corrections:")
    if args.light_correction_off:
        logger.debug("Lighting correction turned on.")
    else:
        logger.debug("Lighting correction turned off.")

    logger.debug("Analysis:")
    if args.endpoint:
        logger.debug("Analysing only last image in series.")
    else:
        logger.debug("Analysing the entire set of images in series.")


def get_grid_format(given_format):
    """Check that the input variable --gridformat is correct.
    Raise an error message if --gridformat is wrong.

    Extract number of rows and columns of the grid by using --gridformat.
    """
    error = False
    # If grid format is provided as nrow x ncol, separate the integers
    if len(given_format) == 1 and len(given_format[0].split("x")) == 2:
        tmp = given_format[0].split("x")
        # Make sure that digits are provided and are higher than 0
        if tmp[0].isdigit() and tmp[1].isdigit():  # Check that inputs are int
            if int(tmp[0]) > 0 and int(tmp[1]) > 0:  # Check that inputs > 0
                nrow, ncol = int(tmp[0]), int(tmp[1])
            else:
                error = True
        else:  # If grid format is not numeric or not separated by x, error
            error = True
    # If grid format is provided as 2 integers showing nrow ncol, use them
    elif len(given_format) == 2:  # Check that inputs are integers and > 0
        if given_format[0].isdigit() and given_format[1].isdigit():
            if int(given_format[0]) > 0 and int(given_format[1]) > 0:
                nrow, ncol = [int(x) for x in given_format]
            else:  # If inputs are not bigger than 0, error
                error = True
        else:  # If grid format is not numeric, error
            error = True
    # If grid format is wrong because there are too many dimensions, error
    else:
        error = True

    # If there has been an error, print it and exit
    if error:
        raise ValueError("""Wrong dimensions for a rectangular grid format.
        Check variable --gridformat and modify it accordingly.""")
    return nrow, ncol


def range_float(x):
    """Check the float is in [0, 1] range."""
    x = float(x)
    if x <= 0.0 or x >= 1.0:
        raise argparse.ArgumentTypeError("%r not in range (0.0, 1.0)" % (x, ))
    return x
